# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band PCA.

This tool allows for the calculation of Principal Components Analysis on
input bands, producing rasters corresponding to the principal components.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> catalog = rs.bandset_catalog()
    >>> # create a BandSets
    >>> file_list_1 = ['file1_b1.tif', 'file1_b2.tif', 'file1_b3.tif']
    >>> catalog.create_bandset(file_list_1, bandset_number=1)
    >>> # start the process
    >>> output = rs.band_pca(input_bands=catalog.get_bandset(1),
    ... output_path='directory_path')
"""

import io
from typing import Union, Optional

import numpy as np

from remotior_sensus.core import (
    configurations as cfg, messages, table_manager as tm
)
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    bands_covariance, calculate_pca, raster_pixel_count
)
from remotior_sensus.util import (
    shared_tools, files_directories, read_write_files
)


def band_pca(
        input_bands: Union[list, int, BandSet],
        output_path: Union[list, str] = None,
        overwrite: Optional[bool] = False,
        nodata_value: Optional[int] = None, extent_list: Optional[list] = None,
        n_processes: Optional[int] = None, available_ram: int = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        number_components: Optional[int] = None,
        progress_message: Optional[bool] = True
) -> OutputManager:
    """Calculation of Principal Components Analysis.

    This tool allows for the calculation of Principal Components Analysis of
    raster bands obtaining the principal components.
    A new raster file is created for each component.

    Args:
        input_bands: input of type BandSet or list of paths or
            integer number of BandSet.
        output_path: string of output path directory or list of paths.
        overwrite: if True, output overwrites existing files.
        nodata_value: value to be considered as nodata.
        extent_list: list of boundary coordinates left top right bottom.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: optional type BandSetCatalog for BandSet number.
        number_components: defines the maximum number of components calculated.
        progress_message: if True then start progress message, if False does not start the progress message (useful if launched from other tools).

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list

    Examples:
        Perform the PCA on the first BandSet
            >>> # import Remotior Sensus and start the session
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> catalog = rs.bandset_catalog()
            >>> # create a BandSets
            >>> file_list_1 = ['file1_b1.tif', 'file1_b2.tif', 'file1_b3.tif']
            >>> catalog.create_bandset(file_list_1, bandset_number=1)
            >>> # start the process
            >>> pca = rs.band_pca(input_bands=catalog.get_bandset(1),output_path='directory_path')
    """  # noqa: E501
    if progress_message:
        cfg.logger.log.info('start')
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
    # prepare process files
    (input_raster_list, raster_info, nodata_list, name_list, warped, out_path,
     vrt_r, vrt_path, n_processes,
     output_list, vrt_list) = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, bandset_catalog=bandset_catalog,
        box_coordinate_list=extent_list
    )
    if number_components is None or number_components > len(input_raster_list):
        number_components = len(input_raster_list)
    # list of band order
    band_order = list(range(len(input_raster_list)))
    cfg.multiprocess.run_separated(
        raster_path_list=input_raster_list, function=raster_pixel_count,
        function_argument=band_order, use_value_as_nodata=nodata_value,
        any_nodata_mask=True, n_processes=n_processes,
        available_ram=available_ram, keep_output_argument=True,
        progress_message='pixel count', min_progress=1, max_progress=20
    )
    cfg.multiprocess.get_dictionary_sum()
    band_dict = cfg.multiprocess.output
    # calculate mean
    max_pixel_count = 0
    try:
        for band_number in band_order:
            band_dict[
                'mean_%s' % band_number] = \
                band_dict['sum_%s' % band_number] / band_dict[
                    'count_%s' % band_number]
            max_pixel_count = max(
                max_pixel_count, band_dict['count_%s' % band_number]
            )
    except Exception as err:
        cfg.logger.log.error(str(err))
        cfg.messages.error(str(err))
        return OutputManager(check=False)
    # calculate covariance
    # dummy bands for memory calculation
    dummy_bands = 2
    cfg.multiprocess.run(
        raster_path=vrt_path, function=bands_covariance,
        function_argument=band_order,
        function_variable=band_dict, use_value_as_nodata=nodata_value,
        any_nodata_mask=True, dummy_bands=dummy_bands,
        keep_output_argument=True, virtual_raster=vrt_r,
        progress_message='calculate covariance',
        min_progress=21, max_progress=40
    )
    cfg.multiprocess.get_dictionary_sum()
    cov_dict = cfg.multiprocess.output
    # calculate covariance matrix
    cov_matrix = np.zeros((len(input_raster_list), len(input_raster_list)))
    # iterate bands
    for _x in band_order:
        for _y in band_order:
            # calculate covariance as [SUM((x - Mean_x) * (y - Mean_y))] / (
            # count - 1)
            cov_matrix[_x, _y] = cov_dict['cov_%s-%s' % (_x, _y)] / (
                    max_pixel_count - 1)
    # calculate sorted eigenvalues and eigenvectors
    eigenvalues, vect = np.linalg.eigh(cov_matrix)
    sorted_args = np.argsort(eigenvalues)
    components = []
    total_variance = []
    cumulative_total_variance = []
    sorted_eigenvalues = []
    for i in reversed(sorted_args):
        if len(components) > number_components:
            break
        sorted_eigenvalues.append(eigenvalues[i])
        components.append(vect[:, i])
        total_variance.append(eigenvalues[i] / eigenvalues.sum() * 100)
        cumulative_total_variance.append(np.sum(total_variance))
    # calculate correlation matrix
    correlation_matrix = np.corrcoef(cov_matrix)
    # create array of means for band normalization
    mean_array = np.zeros((len(input_raster_list)))
    for band_number in band_order:
        try:
            mean_array[band_number] = float(
                band_dict['mean_%s' % str(band_number)]
            )
        except Exception as err:
            cfg.logger.log.error(str(err))
    # run calculation for each component
    output_raster_path_list = []
    min_p = 41
    max_p = int((99 - 1) / number_components)
    for c in range(number_components):
        # check output path
        if vrt_r:
            extension = cfg.vrt_suffix
        else:
            extension = cfg.tif_suffix
        output = shared_tools.join_path(
            files_directories.parent_directory(out_path), '{}_pc{}{}'.format(
                files_directories.file_name(out_path, suffix=False), (c + 1),
                extension
            )
        ).replace('\\', '/')
        cfg.multiprocess.run(
            raster_path=vrt_path, function=calculate_pca,
            function_argument=components[c],
            function_variable=mean_array, calculation_datatype=np.float32,
            use_value_as_nodata=nodata_value, any_nodata_mask=True,
            output_raster_path=output,
            output_data_type=cfg.float32_dt, compress=cfg.raster_compression,
            virtual_raster=vrt_r,
            progress_message='calculating raster component %s' % str(c + 1),
            min_progress=min_p + max_p * c,
            max_progress=min_p + max_p * (c + 1)
        )
        output_raster_path_list.append(output)
    # create principal component table
    try:
        table = _pca_table(
            covariance_matrix=cov_matrix,
            correlation_matrix=correlation_matrix, eigenvectors=components,
            eigenvalues=sorted_eigenvalues, total_variance=total_variance,
            cumulative_total_variance=cumulative_total_variance
        )
    except Exception as err:
        cfg.logger.log.error(str(err))
        cfg.messages.error(str(err))
        return OutputManager(check=False)
    # save principal component details to table
    tbl_out = shared_tools.join_path(
        files_directories.parent_directory(out_path), '{}{}'.format(
            files_directories.file_name(out_path, suffix=False), cfg.csv_suffix
        )
    ).replace('\\', '/')
    read_write_files.write_file(table, tbl_out)
    output_raster_path_list.append(tbl_out)
    if len(output_raster_path_list) == 0:
        cfg.logger.log.error('unable to process files')
        cfg.messages.error('unable to process files')
        return OutputManager(check=False)
    else:
        for i in output_raster_path_list:
            if not files_directories.is_file(i):
                cfg.logger.log.error('unable to process file: %s' % str(i))
                cfg.messages.error('unable to process file: %s' % str(i))
                return OutputManager(check=False)
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; pca: %s' % str(output_raster_path_list))
    return OutputManager(paths=output_raster_path_list)


def _pca_table(
        covariance_matrix, correlation_matrix, eigenvectors, eigenvalues,
        total_variance,
        cumulative_total_variance
):
    """Creates text for table."""
    cv = cfg.comma_delimiter
    nl = cfg.new_line
    text = ['Principal Components Analysis', nl, nl]
    # prepare fields
    dtype_list = []
    fields = ['Bands']
    # column names
    columns = ['Bands', cv]
    vector_columns = ['Bands', cv]
    band_numbers = list(range(1, covariance_matrix.shape[0] + 1))
    for i in band_numbers:
        dtype_list.append((str(i), 'float64'))
        fields.append(str(i))
        columns.append(str(i))
        columns.append(cv)
        vector_columns.append('Vector_%s' % str(i))
        vector_columns.append(cv)
    # remove last cv
    columns.pop(-1)
    vector_columns.pop(-1)
    columns.append(nl)
    vector_columns.append(nl)
    # prepare column text
    columns_text = ''.join(columns)
    vector_columns_text = ''.join(vector_columns)
    # get covariance
    covariance_table = np.rec.fromarrays(covariance_matrix, dtype=dtype_list)
    covariance_table = tm.append_field(
        covariance_table, 'Bands', np.array(band_numbers), 'int8'
    )
    covariance_table = tm.redefine_matrix_columns(
        matrix=covariance_table, input_field_names=fields,
        output_field_names=fields
    )
    # export table creating stream handler
    stream = io.StringIO()
    np.savetxt(stream, covariance_table, delimiter=cv, fmt='%1.4f')
    covariance_text = stream.getvalue()
    text.append('Covariance matrix')
    text.append(nl)
    text.append(columns_text)
    text.append(covariance_text)
    # get correlation
    correlation_table = np.rec.fromarrays(correlation_matrix, dtype=dtype_list)
    correlation_table = tm.append_field(
        correlation_table, 'Bands', np.array(band_numbers), 'int8'
    )
    correlation_table = tm.redefine_matrix_columns(
        matrix=correlation_table, input_field_names=fields,
        output_field_names=fields
    )
    # export table creating stream handler
    stream_2 = io.StringIO()
    np.savetxt(stream_2, correlation_table, delimiter=cv, fmt='%1.4f')
    correlation_text = stream_2.getvalue()
    text.append(nl)
    text.append('Correlation matrix')
    text.append(nl)
    text.append(columns_text)
    text.append(correlation_text)
    # get eigenvectors
    eigenvectors_array = np.zeros(covariance_matrix.shape)
    for v in range(len(eigenvectors)):
        eigenvectors_array[v, ::] = eigenvectors[v]
    eigenvectors_table = np.rec.fromarrays(
        eigenvectors_array, dtype=dtype_list
    )
    eigenvectors_table = tm.append_field(
        eigenvectors_table, 'Bands', np.array(band_numbers), 'int8'
    )
    eigenvectors_table = tm.redefine_matrix_columns(
        matrix=eigenvectors_table, input_field_names=fields,
        output_field_names=fields
    )
    # export table creating stream handler
    stream_3 = io.StringIO()
    np.savetxt(stream_3, eigenvectors_table, delimiter=cv, fmt='%1.4f')
    eigenvectors_text = stream_3.getvalue()
    text.append(nl)
    text.append('Eigenvectors')
    text.append(nl)
    text.append(vector_columns_text)
    text.append(eigenvectors_text)
    # get eigenvalues
    eigenvalues_text = []
    for e in range(len(eigenvalues)):
        eigenvalues_text.append(
            '%1.4f%s%1.4f%s%1.4f%s'
            % (eigenvalues[e], cv, total_variance[e], cv,
               cumulative_total_variance[e], nl)
        )
    text.append(nl)
    text.append('Eigenvalues')
    text.append(cv)
    text.append('Accounted variance')
    text.append(cv)
    text.append('Cumulative variance')
    text.append(nl)
    text.append(''.join(eigenvalues_text))
    joined_text = ''.join(text)
    return joined_text.replace('.0000', '')



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Vector to raster.

This tool allows for the conversion from vector polygons to raster.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> vector = rs.vector_to_raster(vector_path='file.gpkg',
    ...     output_path='vector.tif')
"""  # noqa: E501

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import files_directories, shared_tools, raster_vector


def vector_to_raster(
        vector_path, align_raster: Union[str, BandSet, int],
        vector_field: Optional[str] = None,
        constant: Optional[int] = None,
        pixel_size: Optional[int] = None,
        output_path: Optional[str] = None,
        method: Optional[str] = None,
        area_precision: Optional[int] = 20, resample='mode',
        nodata_value: Optional[int] = 255,
        minimum_extent: Optional[bool] = True,
        extent_list: Optional[list] = None, output_format='GTiff',
        compress=None, compress_format=None,
        n_processes: Optional[int] = None, available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
) -> OutputManager:
    """Performs the conversion from vector to raster.

    This tool performs the conversion from vector polygons to raster.

    Args:
        vector_path: path of vector used as input.
        align_raster: optional string path of raster used for aligning output pixels and projections; it can also be a BandSet or an integer number of a BandSet in a Catalog.
        output_path: string of output path.
        vector_field: the name of the field used as reference value.
        constant: integer value used as reference for all the polygons.
        pixel_size: size of pixel of output raster.
        minimum_extent: if True, raster has the minimum vector extent; if False, the extent is the same as the align raster.
        extent_list: list of boundary coordinates left top right bottom.
        output_format: output format, default GTiff
        method: method of conversion, default pixel_center, other methods are all_touched for burning all pixels touched or area_based for burning values based on area proportion.
        area_precision: for area_based method, the higher the value, the more is the precision in area proportion calculation.
        resample: type for resample when method is area_based.
        compress: if True, compress the output raster.
        compress_format: compress format.
        nodata_value: value to be considered as nodata.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: BandSetCatalog object.

    Returns:
        object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform the conversion to raster of a vector
            >>> vector_to_raster(vector_path='file.gpkg',output_path='vector.tif')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    vector_path = files_directories.input_path(vector_path)
    if type(align_raster) is str:
        input_bands = [align_raster]
    else:
        input_bands = align_raster
    cfg.logger.log.debug('input_bands: %s' % str(input_bands))
    # prepare process files
    (input_raster_list, raster_info, nodata_list, name_list, warped,
     out_path_x, vrt_rx, reference_path, n_processes_x,
     output_list_x, vrt_list_x) = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path,
        n_processes=n_processes, box_coordinate_list=extent_list,
        bandset_catalog=bandset_catalog
    )
    # prepare output
    temp_path = cfg.temp.temporary_file_path(name_suffix=cfg.vrt_suffix)
    if n_processes is None:
        n_processes = cfg.n_processes
    # perform conversion
    if compress is None:
        compress = cfg.raster_compression
    if compress_format is None:
        compress_format = 'DEFLATE21'
    if pixel_size is None:
        x_y_size = None
    else:
        x_y_size = [pixel_size, pixel_size]
    if vector_field is None and constant is None:
        constant = 1
    nodata_value_set = nodata_value
    min_progress = 1
    if method is None or method.lower() == 'pixel_center':
        all_touched = None
        area_based = None
        max_progress = 100
    elif method.lower() == 'all_touched':
        all_touched = True
        area_based = None
        max_progress = 100
    elif method.lower() == 'area_based':
        all_touched = None
        area_based = True
        compress = True
        minimum_extent = False
        nodata_value_set = None
        max_progress = 50
    else:
        all_touched = None
        area_based = None
        max_progress = 100
    cfg.progress.update(message='processing', step=1)
    # open input with GDAL
    cfg.logger.log.debug('vector_path: %s' % vector_path)
    vector_crs = raster_vector.get_crs(vector_path)
    reference_crs = raster_vector.get_crs(reference_path)
    # check crs
    same_crs = raster_vector.compare_crs(vector_crs, reference_crs)
    cfg.logger.log.debug('same_crs: %s' % str(same_crs))
    if not same_crs:
        input_vector = cfg.temp.temporary_file_path(
            name_suffix=files_directories.file_extension(vector_path)
        )
        vector_path = raster_vector.reproject_vector(
            vector_path, input_vector, input_epsg=vector_crs,
            output_epsg=reference_crs
        )
    # perform conversion
    cfg.multiprocess.multiprocess_vector_to_raster(
        vector_path=vector_path, field_name=vector_field,
        output_path=temp_path, reference_raster_path=reference_path,
        output_format=output_format, nodata_value=nodata_value_set,
        background_value=nodata_value, burn_values=constant,
        compress=compress, compress_format=compress_format,
        x_y_size=x_y_size, all_touched=all_touched, area_based=area_based,
        available_ram=available_ram, minimum_extent=minimum_extent,
        area_precision=area_precision, min_progress=min_progress,
        max_progress=max_progress
    )
    if output_path is None:
        output_path = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
    output_path = files_directories.output_path(output_path, cfg.tif_suffix)
    files_directories.create_parent_directory(output_path)
    # resample raster
    if method is not None and method.lower() == 'area_based':
        (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
         scale_offset, data_type) = raster_vector.raster_info(temp_path)
        # copy raster
        left = gt[0]
        top = gt[3]
        p_x_size = gt[1]
        p_y_size = abs(gt[5])
        right = gt[0] + gt[1] * xy_count[0]
        bottom = gt[3] + gt[5] * xy_count[1]
        if compress_format == 'DEFLATE21':
            compress_format = 'DEFLATE -co PREDICTOR=2 -co ZLEVEL=1'
        extra_params = ' -te %s %s %s %s -tr %s %s' % (
            left, bottom, right, top,
            p_x_size * area_precision, p_y_size * area_precision)
        min_progress = 51
        max_progress = 100
        raster_vector.gdal_warping(
            input_raster=temp_path, output=output_path, output_format='GTiff',
            resample_method=resample, compression=True,
            compress_format=compress_format, additional_params=extra_params,
            n_processes=n_processes, dst_nodata=nodata_value,
            min_progress=min_progress, max_progress=max_progress)
    else:
        if files_directories.is_file(temp_path):
            files_directories.move_file(
                in_path=temp_path, out_path=output_path
            )
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; output_path: %s' % output_path)
    return OutputManager(path=output_path)



import datetime
from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestBandSet(TestCase):

    def test_create(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        # create BandSet Catalog
        catalog = rs.bandset_catalog()
        cfg.logger.log.debug('test')
        data_directory = './data/S2_2020-01-01'
        file_list = files_directories.files_in_directory(
            data_directory, sort_files=True, path_filter='S',
            suffix_filter=cfg.tif_suffix
        )
        file_list.pop(0)
        band_names = []
        wavelengths = []
        multiplicative_factors = []
        additive_factors = []
        for f in range(len(file_list)):
            raster_name = files_directories.file_name(file_list[f], False)
            band_names.append(raster_name)
            wavelengths.append(f + 1)
            multiplicative_factors.append(f + 1)
            additive_factors.append(f)
        date = '2020-01-01'
        cfg.logger.log.debug('>>> test bandset create')
        bandset = rs.bandset.create(
            file_list, band_names=band_names, wavelengths=wavelengths,
            multiplicative_factors=multiplicative_factors,
            additive_factors=additive_factors, dates=date, catalog=catalog
        )
        # find bandset values in list
        bandset_values = bandset.find_values_in_list(
            attribute='wavelength', value_list=wavelengths[2:5],
            output_attribute='path'
        )
        self.assertEqual(bandset_values, file_list[2:5])
        self.assertGreater(len(str(bandset.crs)), 0)
        # get BandSet attributes function
        bs = bandset.get_band_attributes
        # BandSet band count
        bs_count = len(bandset.bands)
        self.assertEqual(bs_count, 11)
        # BandSet absolute paths
        bs_apaths = bandset.get_absolute_paths()
        self.assertEqual(bs_apaths, bs('absolute_path'))
        # BandSet relative paths
        bs_paths = bandset.get_paths()
        self.assertEqual(bs_paths, file_list)
        self.assertEqual(bs('path'), file_list)
        # BandSet date
        bs_date = bandset.date
        self.assertEqual(str(bs_date), date)
        self.assertEqual(
            bs('date')[0], datetime.datetime.strptime(date, '%Y-%m-%d').date()
        )
        # multiplicative and additive factors
        bs_multi = bs('multiplicative_factor')
        self.assertEqual(bs_multi, multiplicative_factors)
        bs_add = bs('additive_factor')
        self.assertEqual(bs_add, additive_factors)
        cfg.logger.log.debug('>>> test bandset satellite')
        bandset = rs.bandset.create(
            [data_directory, 'tif'], wavelengths=['Sentinel-2'],
            dates=cfg.date_auto, catalog=catalog
        )
        # BandSet date
        self.assertEqual(str(bandset.date), date)
        # unit
        self.assertEqual(bandset.get_wavelength_units()[0], cfg.wl_micro)
        # spectral range bands
        (blue_band, green_band, red_band, nir_band, swir_1_band,
         swir_2_band) = bandset.spectral_range_bands(output_as_number=False)
        self.assertAlmostEqual(blue_band['wavelength'], cfg.blue_center, 1)
        self.assertAlmostEqual(red_band['wavelength'], cfg.red_center, 1)
        self.assertAlmostEqual(nir_band['wavelength'], cfg.nir_center, 1)
        self.assertEqual(
            blue_band['wavelength_unit'], bandset.get_wavelength_units()[0]
        )
        self.assertGreater(len(red_band['crs']), 0)
        self.assertGreater(len(nir_band['absolute_path']), 0)
        """
        cfg.logger.log.debug('>>> test tools')
        expression = (
                cfg.variable_band_quotes + cfg.variable_ndvi_name
                + cfg.variable_band_quotes
        )
        calc = bandset.calc(expression)
        self.assertTrue(calc.check)
        self.assertTrue(files_directories.is_file(calc.paths[0]))
        combination = bandset.combination()
        self.assertTrue(combination.check)
        self.assertTrue(files_directories.is_file(combination.paths[0]))
        dilation = bandset.dilation(value_list=[1000], size=2)
        self.assertTrue(dilation.check)
        self.assertTrue(files_directories.is_file(dilation.paths[0]))
        erosion = bandset.erosion(value_list=[1000], size=1)
        self.assertTrue(erosion.check)
        self.assertTrue(files_directories.is_file(erosion.paths[0]))
        pca = bandset.pca()
        self.assertTrue(pca.check)
        self.assertTrue(files_directories.is_file(pca.paths[0]))
        sieve = bandset.sieve(size=3, connected=True)
        self.assertTrue(sieve.check)
        self.assertTrue(files_directories.is_file(sieve.paths[0]))
        neighbor_pixels = bandset.neighbor_pixels(
            size=10, circular_structure=True, stat_name='Sum'
        )
        self.assertTrue(neighbor_pixels.check)
        self.assertTrue(files_directories.is_file(neighbor_pixels.paths[0]))
        """
        cfg.logger.log.debug('>>> test methods')
        # get band by nearest wavelength
        band_x = bandset.get_band_by_wavelength(
            cfg.green_center, threshold=0.1
        )
        self.assertEqual(band_x['wavelength'], cfg.green_center)
        # get bands by attribute
        band_b = bandset.get_bands_by_attributes(
            'wavelength', band_x['wavelength']
        )
        self.assertEqual(band_b['wavelength'], band_x['wavelength'])
        file_list = ['S2_2020-01-01/S2_B01.tif', 'S2_2020-01-01/S2_B02.tif',
                     'S2_2020-01-01/S2_B03.tif']
        cfg.logger.log.debug('>>> test bandset dates')
        bandset = rs.bandset.create(
            file_list, wavelengths=['Sentinel-2'], dates=cfg.date_auto,
            root_directory='./data', catalog=catalog
        )
        # get BandSet attributes function
        bs = bandset.get_band_attributes
        # BandSet names
        bs_names = bs('name')
        self.assertEqual(
            bs_names[0], files_directories.file_name(file_list[0], False)
        )
        # BandSet relative paths
        bs_paths = bandset.get_paths()
        self.assertEqual(bs_paths[0], file_list[0])
        # bandset wavelengths
        bs_wls = bandset.get_wavelengths()
        self.assertEqual(bs_wls, bs('wavelength'))
        self.assertEqual(bandset.get_wavelengths(), bs('wavelength'))
        self.assertEqual(
            bandset.get_band_attributes('wavelength')[0],
            cfg.satellites[cfg.satSentinel2][0][0]
        )
        # get BandSet band function
        band = bandset.get_band
        # band wavelength
        b_wl = band(1)['wavelength']
        self.assertEqual(b_wl, cfg.satellites[cfg.satSentinel2][0][0])
        # band absolute path
        b_apath = str(band(1)['absolute_path'])
        self.assertEqual(
            files_directories.file_name(b_apath),
            files_directories.file_name(file_list[0], False)
        )
        # band x size
        b_x_size = band(1).x_size
        self.assertGreater(b_x_size, 0)
        # band nodata
        b_nodata = band(1).nodata
        self.assertGreater(b_nodata, 0)
        # reset
        bandset.reset()
        self.assertEqual(bandset.bands, None)
        # sort bands by wavelength
        wavelengths = list(range(1, len(file_list) + 1))
        wavelengths.pop(-1)
        wavelengths.pop(-1)
        wavelengths.append(3)
        wavelengths.append(2)
        bandset_w = rs.bandset.create(
            file_list, wavelengths=wavelengths, root_directory='./data',
            catalog=catalog
        )
        bandset_w.sort_bands_by_wavelength()
        self.assertEqual(bandset_w.get_band_attributes('wavelength')[-1], 3)
        # box coordinate list
        coordinate_list = [230250, 4674550, 230320, 4674440]
        bandset = rs.bandset.create(
            file_list, wavelengths=['Sentinel-2'], dates=cfg.date_auto,
            root_directory='./data', box_coordinate_list=coordinate_list,
            catalog=catalog
        )
        self.assertEqual(bandset.box_coordinate_list, coordinate_list)

        # clear temporary directory
        rs.close()
